/**
 *  Regularizer that penalizes features in proportion to 
 *  the distances of the alternations that they force according 
 *  to the quantified P-map (Steriade 1997, 2001).
 */

import java.util.*;
import cern.colt.matrix.*;

public class PMapRegularizer implements Regularizer
{
double MU = 0.0;
double SIGMA = 1.0;
HashMap PMap;

PMapRegularizer()
{
	// Hard-wired P-Map for the velar palatalization study
	PMap = new HashMap();
	PMap.put("K IY", new Double(9.23));			// ki -> tSi
	PMap.put("K EY", new Double(12.68));		// ke -> tSe		// checked 05/05/05
	PMap.put("K AA", new Double(88.72));		// ka -> tSa
	PMap.put("G IY", new Double(21.13));		// gi -> dZi
	PMap.put("G EY", new Double(40.60));		// ge -> dZe		// checked 05/05/05
	PMap.put("G AA", new Double(126.93));		// ga -> dZa
}


public double value(Feature[] features, double[] weights)
{
	double val = 0.0;

	int K = features.length;
	for(int k = 0; k < K; k++)
	{
		double sigma_k = PMapCost(features[k]);
		val += Math.pow(weights[k] - MU, 2.0)/(2 * Math.pow(1/sigma_k,2.0));
	}

	return val;
}

public double[] gradient(Feature[] features, double[] weights)
{ 
	int K = features.length;
	double[] grad = new double[K];

	for(int k = 0; k < K; k++)
	{
		double sigma_k = PMapCost(features[k]);
		grad[k] = (weights[k] - MU)/(Math.pow(1/sigma_k,2.0));
	}
	return grad;
}

public double PMapCost(Feature f)
{
	// faithfulness features have default sigma^2
	if(!(f instanceof MarkednessFeature)) return SIGMA;

	// determine P-map cost of satisfiying a
	// markedness feature by palatalizing
	String ftr = ((MarkednessFeature) f).matcher.pattern().pattern();
	String[] ftr_segs = ftr.split(" ");
	String cost = "";

	if(ftr_segs[0].indexOf("G") != -1)
	{ cost = cost + "G"; }
	else
	{ cost = cost + "K"; }
	
	if(ftr_segs[1].indexOf("AA") != -1)
	{ cost = cost + " AA"; }
	else if(ftr_segs[1].indexOf("EY") != -1)
	{ cost = cost + " EY"; }
	else
	{ cost = cost + " IY"; }
	
//xxx testing only	System.out.println(((MarkednessFeature) f).matcher.pattern().pattern() + 
//xxx		"\t" + ((Double) PMap.get(cost)).doubleValue());
	
	double sigma_f =  ((Double) PMap.get(cost)).doubleValue();

// xxx crude bias for more general features: 10 'pts' added to 1/sigma
// for every natural class in the feature
/*
	for(int i = 0; i < ftr_segs.length; i++)
	{
		if(ftr_segs[i].equals("")) continue;
		if(ftr_segs[i].equals("((IY)|(EY)|(AA))")) continue;	// longhand for 'any segment'
		sigma_f += 10.0;
	}
	
*/
	return sigma_f;
}

public double sigma2(Feature feature)
{ return Math.pow(PMapCost(feature), 2.0); }

}